using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Multiple precision Complex Matrix algorithms
	///		in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	// ------- A multiple-precision signed floating matrix
	public class MultiZ : MultiC	// array of complex
	{
		public int Rs;		// Number of rows
		public int Cs;		// Number of columns
		public MultiC[,] E;	// elements

		public MultiZ()				// initialize a scalar with M = 1
			: base(0)				// calls MultiC(0)
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiC[0, 0];
		}

		public MultiZ(int nMaxI)		// initialize a scalar with M = nMaxI
			: base(nMaxI)				// calls MultiC(i)
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiC[0, 0];
		}

		public MultiZ(int nMaxI, int rs, int cs)	// initialize a matrix
			: base(0)								// calls MultiC(0)
		{
			// constructor logic
			if (rs < 1 || cs < 1)
			{
				rs = cs = 0;						// change to a scalar
				this.V = new int[nMaxI+1];
				this.M = nMaxI;
				this.N = 1;
				this.I.V = new int[nMaxI+1];
				this.I.M = nMaxI;
				this.I.N = 1;
			}
			this.Rs = rs;
			this.Cs = cs;
			this.E = new MultiC[rs, cs];
			for (int r = 0; r < rs; r++)
				for (int c = 0; c < cs; c++)
					this.E[r, c] = new MultiC(nMaxI);
		}

		//		public static bool operator = (MultiZ left, MultiZ right)	// not allowed to overload =
//		{
//			MultiZD.SetToZ(right, left);
//			return true;
//		}

		public static MultiZ operator & (MultiZ left, MultiZ right)	// use &= for = assignment
		{
			MultiZD.SetToZ(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, MultiC right)	// use &= for = assignment
		{
			MultiCD.SetToC(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, MultiF right)	// use &= for = assignment
		{
			MultiZD.SetToFZ(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, MultiSI right)	// use &= for = assignment
		{
			MultiZD.SetToSIZ(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, int right)	// use &= for = assignment
		{
			MultiZD.SetTo1Z(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, long right)	// use &= for = assignment
		{
			MultiZD.SetToExZ(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, double right)	// use &= for = assignment
		{
			MultiZD.SetToDbZ(right, left);
			return left;
		}

		public static MultiZ operator & (MultiZ left, decimal right)	// use &= for = assignment
		{
			MultiZD.SetToDeZ(right, left);
			return left;
		}

		public static MultiZ operator + (MultiZ left, MultiZ right) // can be used for a += b
		{
			MultiZ z = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.AddZ(left, right, z);
			return z;
		}

		public static MultiZ operator + (MultiZ left, int right) // can be used for a += b
		{
			MultiZ z = new MultiZ(Calc.fMC);
			MultiZD.Add1Z(left, right, z);
			return z;
		}

		public static MultiZ operator + (int left, MultiZ right) // can be used for a += b
		{
			MultiZ z = new MultiZ(Calc.fMC);
			MultiZD.Add1Z(right, left, z);
			return z;
		}

		public static MultiZ operator - (MultiZ left, MultiZ right) // can be used for a -= b
		{
			MultiZ z = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.SubZ(left, right, z);
			return z;
		}

		public static MultiZ operator - (MultiZ left, int right) // can be used for a -= b
		{
			MultiZ z = new MultiZ(Calc.fMC);
			MultiZD.Add1Z(left, -right, z);
			return z;
		}

		public static MultiZ operator - (int left, MultiZ right) // can be used for a -= b
		{
			MultiZ z = new MultiZ(Calc.fMC);
			MultiZD.Add1Z(right, -left, z);
			MultiZD.ChSZ(z);
			return z;
		}

		public static MultiZ operator - (MultiZ x)
		{
			MultiZ z = new MultiZ(x.N, x.Rs, x.Cs);
			MultiZD.SetToMS(x, z);
			MultiZD.ChSZ(z);
			return z;
		}
		public static MultiZ operator * (MultiZ left, MultiZ right) // can be used for a *= b
		{
			MultiZ z = new MultiZ(Calc.fMC, left.Rs, right.Cs);
			MultiZD.MulZ(left, right, z);
			return z;
		}

		public static MultiZ operator * (MultiZ left, int right) // can be used for a *= b
		{
			MultiZ z = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.Mul1Z(left, right, z);
			return z;
		}

		public static MultiZ operator * (int left, MultiZ right)
		{
			MultiZ z = new MultiZ(Calc.fMC, right.Rs, right.Cs);
			MultiZD.Mul1Z(right, left, z);
			return z;
		}

		public static MultiZ operator / (MultiZ left, MultiZ right) // can be used for a /= b
		{
			int maxRs = Math.Max(left.Rs, right.Rs);
			int maxCs = Math.Max(left.Cs, right.Cs);
			MultiZ q = new MultiZ(Calc.fMC, maxRs, maxCs);
			MultiZD.DivZ(left, right, q);
			return q;
		}

		public static MultiZ operator / (MultiZ left, MultiC right) // can be used for a /= b
		{
			MultiZ q = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.DivZ(left, right, q);
			return q;
		}

		public static MultiZ operator / (MultiZ left, MultiF right) // can be used for a /= b
		{
			MultiZ q = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.DivMF(left, right, q);
			return q;
		}

		public static MultiZ operator / (MultiZ left, int right) // can be used for a /= b
		{
			MultiZ q = new MultiZ(Calc.fMC, left.Rs, left.Cs);
			MultiZD.Div1Z(left, right, q);
			return q;
		}

		// % and ^ will default to Singed Integer versions

		public static bool operator == (int d1, MultiZ z)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return false;
			}
			MultiZ dF = new MultiZ(MultiID.sdInInt);
			MultiZD.SetTo1Z(d1, dF);
			return (MultiFD.CompFF(dF, z) == 0);
		}

		public static bool operator != (int d1, MultiZ z)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return true;
			}
			MultiZ dF = new MultiZ(MultiID.sdInInt);
			MultiZD.SetTo1Z(d1, dF);
			return (MultiFD.CompFF(dF, z) != 0);
		}

		public static bool operator == (MultiZ z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return false;
			}
			MultiZ dF = new MultiZ(MultiID.sdInInt);
			MultiZD.SetTo1Z(d1, dF);
			return (MultiFD.CompFF(dF, z) == 0);
		}

		public static bool operator != (MultiZ z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return true;
			}
			MultiZ dF = new MultiZ(MultiID.sdInInt);
			MultiZD.SetTo1Z(d1, dF);
			return (MultiFD.CompFF(dF, z) != 0);
		}

		public static bool operator == (long ex, MultiZ z)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return false;
			}
			MultiZ exF = new MultiZ(MultiID.sdInLong);
			MultiZD.SetToExZ(ex, exF);
			return (MultiFD.CompFF(exF, z) == 0);
		}

		public static bool operator != (long ex, MultiZ z)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return true;
			}
			MultiZ exF = new MultiZ(MultiID.sdInLong);
			MultiZD.SetToExZ(ex, exF);
			return (MultiFD.CompFF(exF, z) != 0);
		}

		public static bool operator == (MultiZ z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return false;
			}
			MultiZ exF = new MultiZ(MultiID.sdInLong);
			MultiZD.SetToExZ(ex, exF);
			return (MultiFD.CompFF(exF, z) == 0);
		}

		public static bool operator != (MultiZ z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiZD.ErrIncompatible();
				return true;
			}
			MultiZ exF = new MultiZ(MultiID.sdInLong);
			MultiZD.SetToExZ(ex, exF);
			return (MultiFD.CompFF(exF, z) != 0);
		}

		public static bool operator == (MultiZ x, MultiZ y)
		{
			return MultiZD.EqualZ(x, y);
		}

		public static bool operator != (MultiZ x, MultiZ y)
		{
			return !MultiZD.EqualZ(x, y);
		}

		public static bool operator < (MultiZ x, MultiZ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) < 0);
		}

		public static bool operator > (MultiZ x, MultiZ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) > 0);
		}

		public static bool operator <= (MultiZ x, MultiZ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) <= 0);
		}

		public static bool operator >= (MultiZ x, MultiZ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) >= 0);
		}

		public static MultiZ operator ++ (MultiZ x)
		{
			if (x.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return MultiZD.mZero;
			}
			MultiFD.RAddF(MultiFD.fOne, x);
			return x;
		}

		public static MultiZ operator -- (MultiZ x)
		{
			if (x.Rs != 0)
			{
				MultiZD.ErrNotScalar();
				return MultiZD.mZero;
			}
			MultiFD.RAddF(MultiFD.fMOne, x);
			return x;
		}

		public override bool Equals(object o)	// dummy
		{
			return true;
		}

		public override int GetHashCode()		// dummy
		{
			return 0;
		}
	}

	public class MultiZD
	{
		public MultiZD()	// constructor logic
		// initialize Multiple-precision floating matrix package
		// input: MultiID.muDMax, MultiID.muNMax, MultiID.baseM
		{
			mZero = new MultiZ(1);			// mZero = 0.0
			mOne = new MultiZ(1);
			SetTo1Z(1, mOne);				// mOne = 1.0
			mTwo = new MultiZ(1);
			SetTo1Z(2, mTwo);				// mTwo = 2.0
			vZero = new MultiZ(1, 1, 1);	// vZero = {0.0}
		}

		public static bool EqualZ(MultiZ x, MultiZ y)
		{
			if (x.Rs == 0 && y.Rs == 0)
				return (MultiFD.CompFF(y, x) == 0);
			if (x.Rs != y.Rs || x.Cs != y.Cs)
			{
				ErrIncompatible();
				return false;
			}
			bool eq;
			for (int r = 0; r < x.Rs; r++)
				for (int c = 0; c < x.Cs; c++)
				{
					eq = (MultiFD.CompFF(x.E[r, c], y.E[r, c]) == 0);
					if (!eq)
						return false;
				}
			return true;
		} // EqualM

		public static void Matrix(int nMaxI, int rs, int cs, out MultiZ z) // init matrix
		{
			z = new MultiZ(nMaxI, rs, cs);
		}

		public static void DoneZ(ref MultiZ z) // free up space
		{
			z = new MultiZ();
		}

		public static void ClearZ(ref MultiZ z)	// set Value = 0
		{
			z = new MultiZ(z.M, z.Rs, z.Cs);
		}

		public static void ChSZ(MultiZ z) // change sign, z = - z
		{
			MultiCD.ChSC(z);
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)
					MultiCD.ChSC(z.E[r, c]);
		}

		public static void SetToZ(MultiZ x, MultiZ z)	// z = x
			// can reduce precision and round
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiCD.SetToC(x, z);
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiCD.SetToC(x.E[r, c], z.E[r, c]);
					}
				}
			}
		}

		public static void SetToSIZ(MultiSI x, MultiZ z) // z = x
		// can reduce precision and round
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToSIF(x, z);
		}

		public static void SetToMS(MultiZ x, MultiZ z) // z = x, same size
		// copy everything but .Nm Name
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiCD.SetToCS(x, z);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
						MultiCD.SetToCS(x.E[r, c], z.E[r, c]);
			}
		}

		public static void SetToFZ(MultiF x, MultiZ z)	// z = x
			// can reduce precision and round
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToF(x, z);
		}

		public static void NormZ(MultiZ z) // normalize
		{
			MultiFD.NormF(z);
			for (int r = 0; r < z.Rs; r++)
			{
				for (int c = 0; c < z.Cs; c++)
				{
					MultiFD.NormF(z.E[r, c]);
				}
			}
		}

		public static void RAddZ(MultiZ x, MultiZ z) // z = z + x
		// sets muErr true if overflow
		{
			AddZ(z, x, z);
		}

		public static void RSubZ(MultiZ x, MultiZ z) // z = z - x
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, z))
				ClearZ(ref z);
			else
			{
				ChSZ(x);
				RAddZ(x, z);
				ChSZ(x);
			}		
		}

		public static void AddZ(MultiZ x, MultiZ y, MultiZ z) // z = x + y
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiFD.AddF(x, y, z);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				ErrIncompatible();
			else
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						z.E[r, c] &= x.E[r, c] + y.E[r, c];
		}

		public static void SubZ(MultiZ x, MultiZ y, MultiZ z) // z = x - y
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, y))
				ClearZ(ref z);
			else if (Object.ReferenceEquals(x, z))
				RSubZ(y, z);
			else if (Object.ReferenceEquals(y, z))
			{
				RSubZ(x, z);
				ChSZ(z);
			}
			else
			{
				ChSZ(y);
				AddZ(x, y, z);
				ChSZ(y);
			}
		}

		public static void SetTo1Z(int d1, MultiZ z)	// z = d1
		// allows a literal d1, d1 an integer, d1 not changed
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetTo1F(d1, z);
		}

		public static void SetToExZ(Int64 ext, MultiZ z)	// z = ext (Int64 = long)
		// allows a literal ext, ext not changed, max Int64 = 9223,37203,68547,75807 = 2^63-1
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToExF(ext, z);
		}

		public static void SetToDbZ(double db, MultiZ z) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToDbF(db, z);
		}

		public static void SetToDeZ(decimal de, MultiZ z) // z = db
		// allows a literal de, de not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiC[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToDeF(de, z);
		}

		public static void GetDbInZ(out double db, double low, double high, MultiZ z)
		// db = z % baseM^4, in a range, + or -
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.GetDbInF(out db, low, high, z);
		}

		public static void Add1Z(MultiZ x, int d1, MultiZ z) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (x.Rs != 0 || z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.Add1F(x, d1, z);
		}

		public static void WriteFixedL(RichTextBox outB, StreamWriter outF, MultiZ x, ref int tot, int width)
		{
			if (x.Rs == 0)
			{
				MultiCD.WriteFixedL(outB, outF, x, ref tot, width);
			}
			else
			{
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				MultiID.WriteT(outB, outF, "{", ref tot, width);
				int rs = x.Rs;
				int cs = x.Cs;
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.MuInterrupt();		// allow for abort
						if (MultiID.muAbort) break;	// exit for
						MultiCD.WriteFixedL(outB, outF, x.E[r, c], ref tot, width);
						if (MultiID.diagOn && Calc.listOn)
							MultiID.WriteT(" N=" + x.E[r, c].N + " M=" +
								x.E[r, c].M + " U=" + x.E[r, c].V.GetUpperBound(0), ref tot);
						if (r == rs-1 && c == cs-1)
							MultiID.WriteT(outB, outF, "}", ref tot, width);
						else
						{
							if (c == cs-1)
							{
								MultiID.WriteTLn(outB, outF, ";+", ref tot, width);
								MultiID.WriteTLn(outB, outF, "+", ref tot, width);
							}
							else
								MultiID.WriteTLn(outB, outF, ",+", ref tot, width);
						}
					}
					if (MultiID.muAbort) break;	// exit for
				}
				MultiID.writeBool = wB;
			}
		} // WriteFixedL

		private static void Write1Z(string st, ref int j, int k, ref int did, int didZ,
			RichTextBox outB, StreamWriter outF, ref int tot, int width)
		// part of WriteFixed
		{
			while (j < k)
			{
				j++;
				if (did != didZ && MultiID.muDpG > 0)
				{
					if (did % MultiID.muDpG == 0) MultiID.WriteT(outB, outF, ",", ref tot, width);
				}
				MultiID.WriteT(outB, outF, st.Substring(j - 1, 1), ref tot, width);
				did++;
			}
			return;
		}

		public static void WriteZ(MultiZ z, ref int tot)
		{
			WriteZ(MultiID.outS, null, z, ref tot, 0);
		}

		public static void WriteZ(RichTextBox outB, StreamWriter outF, MultiZ x, ref int tot, int width)
		// output x as a line of text, tot = characters on line so far, x is not modified
		// good for any base = baseO [2, 36], base 10 much faster
		{
			if (x.Rs == 0)
			{
				MultiCD.WriteC(outB, outF, x, ref tot, width);
			}
			else
			{
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				MultiID.WriteT(outB, outF, "{", ref tot, width);
				int rs = x.Rs;
				int cs = x.Cs;
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.MuInterrupt();		// allow for abort
						if (MultiID.muAbort) break;	// exit for
						MultiCD.WriteC(outB, outF, x.E[r, c], ref tot, width);
						if (MultiID.diagOn && Calc.listOn)
						{
							MultiID.WriteT(" N=" + x.E[r, c].N + " M=" + x.E[r, c].M, ref tot);
							MultiID.WriteT(" U=" + x.E[r, c].V.GetUpperBound(0), ref tot);
							MultiID.WriteT(" I.N=" + x.E[r, c].I.N + " I.M=" + x.E[r, c].I.M, ref tot);
							MultiID.WriteT(" I.U=" + x.E[r, c].I.V.GetUpperBound(0), ref tot);
						}
						if (r == rs-1 && c == cs-1)
							MultiID.WriteT(outB, outF, "}", ref tot, width);
						else
						{
							if (c == cs-1)
							{
								MultiID.WriteTLn(outB, outF, ";+", ref tot, width);
								MultiID.WriteTLn(outB, outF, "+", ref tot, width);
							}
							else
								MultiID.WriteTLn(outB, outF, ",+", ref tot, width);
						}
					}
					if (MultiID.muAbort) break;	// exit for
				}
				MultiID.writeBool = wB;
			}
		} // WriteM

		public static void RMul1Z(int d1, MultiZ z) // z = z * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			if (z.Rs == 0)
				MultiFD.RMul1F(d1, z);
			else
				for (int r = 0; r < z.Rs; r++)
					for (int c = 0; c < z.Cs; c++)
						MultiFD.RMul1F(d1, z.E[r, c]);
		}

		public static void Mul1Z(MultiZ x, int d1, MultiZ z) // z = x * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			SetToZ(x, z);
			RMul1Z(d1, z);
		}

		public static MultiZ Diagonal(MultiZ x) // return diagonal matrix from vector x
		{
			if (x.Rs == 0)
				ErrNotMatrix();
			else if (x.Rs != 1 && x.Cs != 1)
				ErrIncompatible();
			else
			{
				int n = Math.Max(x.Rs, x.Cs);
				MultiZ t;
				Matrix(Calc.fMC, n, n, out t);
				for (int r = 0; r < n; r++)
					for (int c = 0; c < n; c++)
						t.E[r, c] &= MultiCD.cZero;
				if (x.Rs == 1)
					for (int i = 0; i < n; i++)
						Calc.SetToNM(x.E[0, i], t.E[i, i]);
				else
					for (int i = 0; i < n; i++)
						Calc.SetToNM(x.E[i, 0], t.E[i, i]);
				return t;
			}
			return mZero;
		} // DiagonalC

		public static void MulZ(MultiZ x, MultiZ y, MultiZ z) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiCD.MulC(x, y, z);
			else if (x.Rs == 0)
			{
				MultiZ zM = new MultiZ(Calc.fMC, y.Rs, y.Cs);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						MultiCD.SetToC(x * y.E[r, c], zM.E[r, c]);
				SetToZ(zM, z);
			}
			else if (y.Rs == 0)
			{
				MultiZ zM = new MultiZ(Calc.fMC, x.Rs, x.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						MultiCD.SetToC(y * x.E[r, c], zM.E[r, c]);
				SetToZ(zM, z);
			}
			else if (x.Cs != y.Rs)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiC pC = new MultiC(Calc.fMC);
				MultiZ zM = new MultiZ(Calc.fMC, x.Rs, y.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
					{
						pC &= 0;
						for (int i = 0; i < x.Cs; i++)
						{
							pC += x.E[r, i] * y.E[i, c];
						}
						zM.E[r, c] &= pC;
					}
				SetToZ(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MulN

		public static void RSqZ(MultiZ z) // z = z * z
		// sets muErr true if overflow
		{
			MulZ(z, z, z);
		}

		public static void RDivMF(MultiF x, MultiZ z) // z = z / x
		// sets muErr true if overflow
		{
			DivMF(z, x, z);
		}

		public static void RDiv1Z(int d1, MultiZ z) // z = z / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiF x = new MultiF(2); // temp
			MultiFD.SetTo1F(d1, x);
			RDivMF(x, z);
		}

		public static void Div1Z(MultiZ u, int d1, MultiZ z) // z = u / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			SetToZ(u, z);
			RDiv1Z(d1, z);
		}

		public static void DivZ(MultiZ u, MultiC d, MultiZ z) // z = u / d (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiZ dInv = new MultiZ(Calc.fMC);
			dInv &= MultiCD.cOne / d;
			MulZ(u, dInv, z);
		} // DivZ

		public static void DivZ(MultiZ u, MultiZ d, MultiZ z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (u.Rs == 0 && d.Rs == 0)
				MultiCD.DivC(u, d, z);
			else if (d.Rs != d.Cs)
			{
				ErrNotSquare();
			}
			else if (u.Rs > 0 && d.Rs > 0 && (u.Rs != u.Cs || u.Rs != d.Rs))
			{
				ErrIncompatible();
			}
			else if (d.Rs == 0)
			{
				MultiZ dInv = new MultiZ(Calc.fMC);
				dInv &= MultiCD.cOne / (MultiC)d;
				MulZ(u, dInv, z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ dInv = new MultiZ(Calc.fMC, d.Rs, d.Cs);
				InvZ(d, dInv);
				MulZ(u, dInv, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				NormZ(z);
			}
		} // DivZ

		public static void InvZ(MultiZ x, MultiZ z) // z = 1 / x
		{
			if (x.Rs == 0)
				MultiFD.DivF(MultiCD.cOne, x, z);
			else if (x.Cs != x.Rs)
				ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int i, j;
				int n = x.Rs;
				MultiZ a = new MultiZ(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiC d = new MultiC();
				bool sing;
				LUdecomp(a, index, d, out sing); // decompose the matrix just once
				if (sing)
					ErrSingular();
				else
				{
					MultiZ y = new MultiZ(Calc.fMC, n, n);
					MultiZ b = new MultiZ(Calc.fMC, n, 1);
					for (j = 0; j < n; j++)			// solve for column j
					{
						for (i = 0; i < n; i++)		// set up identity matrix column j
						{
							if (i == j)
								b.E[i, 0] &= MultiFD.fOne;
							else
								b.E[i, 0] &= MultiFD.fZero;
						}
						LUbackSub(a, index, b);
						for (i = 0; i < n; i++)		// store column j
							y.E[i, j] &= b.E[i, 0];
					}
					SetToZ(y, z);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				NormZ(z);
			}
		} // InvZ
		
		public static void LUdecomp(MultiZ a, int[] index, MultiC d, out bool sing)
		// LU decomposition, sing is singular matrix flag
		// call XPCalLst.Expand(); before calling LUdecomp
// this code is adapted from Numerical Recipes in C
		{
			sing = false;
			int n = a.Rs;
			int nm = n - 1;
			int i, j, k;
			int iMax = 0;
			MultiC big = new MultiC(Calc.fMC);
			MultiC dum = new MultiC(Calc.fMC);
			MultiC sum = new MultiC(Calc.fMC);
			MultiC temp = new MultiC(Calc.fMC);
			MultiC[] vv;
			vv = new MultiC[n];
			for (i  = 0; i < n; i++)
				vv[i] = new MultiC(Calc.fMC);
			d &= MultiCD.cOne;					// no rows interchanged yet
			for (i = 0; i < n; i++)				// loop over rows to get implicit scaling information
			{
				big &= MultiCD.cZero;
				for (j = 0; j < n; j++)
				{
					MultiCD.Abs(a.E[i, j], temp);
					if (temp > big) 
						big &= temp;
				}
				if (big == MultiCD.cZero)
				{
					d &= MultiCD.cZero;			// singular matrix
					sing = true;
					return;
				}
				vv[i] &= MultiCD.cOne / big;	// save the scaling
			}
			for (j = 0; j < n; j++)				// this is the loop over columns of C
			{
				for (i = 0; i < j; i++)
				{
					sum &= a.E[i, j];
					for (k = 0; k < i; k++)
						sum -= a.E[i, k] * a.E[k, j];
					a.E[i, j] &= sum;
				}
				big &= MultiCD.cZero;			// initialize for the search for largest pivot element
				for (i = j; i < n; i++)
				{
					sum &= a.E[i, j];
					for (k = 0; k < j; k++)
						sum -= a.E[i, k] * a.E[k, j];
					a.E[i, j] &= sum;
					MultiCD.Abs(sum * vv[i], dum);	// figure of merit for the pivot
					if (dum >= big)					// is it better than the best so far?
					{
						big &= dum;
						iMax = i;
					}
				}
				if (j != iMax)					// do we need to interchange rows?
				{
					for (k = 0; k < n; k++)		// yes we do
					{
						dum &= a.E[iMax, k];
						a.E[iMax, k] &= a.E[j, k];
						a.E[j, k] &= dum;
					}
					d.S = !d.S;					// and change the parity of d
					vv[iMax] &= vv[j];			// also interchange the scale factor
				}
				index[j] = iMax;
				if (a.E[j, j] == MultiCD.cZero)
				{
					d &= MultiCD.cZero;			// singular matrix
					sing = true;
					return;
				}
				if (j != nm)					// now finally divide by the pivot element
				{
					dum &= MultiCD.cOne / a.E[j, j];	// if the pivot element is zero the matrix is singular
					for (i = j + 1; i < n; i++)			// for some applications on singular matrices it is
						a.E[i, j] *= dum;				// desirable to substitute tiny for zero
				}
			}
			long next;
			long max = Int64.MinValue;
			for (i = 0; i < n; i++)
			{
				next = Int64.MinValue;
				if ((MultiF)a.E[i, i] != MultiFD.fZero)
					next = a.E[i, i].N + a.E[i, i].C;
				if ((MultiF)a.E[i, i].I != MultiFD.fZero)
					next = Math.Max(next, a.E[i, i].I.N + a.E[i, i].I.C);
				if (next > max)
					max = next;
			}
			for (i = 0; i < n; i++)
			{
				next = a.E[i, i].N + a.E[i, i].C;
				if (max - next >= Calc.fMC - MultiFD.fGuard)
					XPCalLst.SetToNM(MultiFD.fZero, a.E[i, i]);
				next = a.E[i, i].I.N + a.E[i, i].I.C;
				if (max - next >= Calc.fMC - MultiFD.fGuard)
					XPCalLst.SetToNM(MultiFD.fZero, a.E[i, i].I);
				if (a.E[i, i] == MultiCD.cZero)
					sing = true;
			}
		} // LUdecomp

		public static void LUbackSub(MultiZ a, int[] index, MultiZ b)
			// this code is adapted from Numerical Recipes in C
		{
			int n = a.Rs;
			int nm = n - 1;
			int i, ip, j;
			int ii = -1;
			MultiC sum = new MultiC(Calc.fMC);
			for (i = 0; i < n; i++) 
			{
				ip = index[i];
				sum &= b.E[ip, 0];
				b.E[ip, 0] &= b.E[i, 0];
				if (ii != -1)
					for (j = ii; j < i; j++)
						sum -= a.E[i, j] * b.E[j, 0];
				else if (sum != MultiCD.cZero)
					ii = i;
				b.E[i, 0] &= sum;
			}
			for (i = nm; i >= 0; i--) 
			{
				sum &= b.E[i, 0];
				for (j = i + 1; j < n; j++)
					sum -= a.E[i, j] * b.E[j, 0];
				b.E[i, 0] &= sum / a.E[i, i];
			}
		} // LUbackSub

		public static void DivMF(MultiZ u, MultiF d, MultiZ z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			z &= (MultiF)u / d;
			z.I &= (MultiF)u.I / d;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			NormZ(z);
		} // DivMF

		public static void LopZ(MultiZ x, MultiZ z) // z = x with one less super digit
		{
			if (x.Rs == 0)
				MultiCD.LopC(x, z);
			else
			{
				long smallC = Int64.MaxValue;
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						if (x.E[r, c].C < smallC && x.E[r, c] != MultiFD.fZero)
							smallC = x.E[r, c].C;
						if (x.E[r, c].I.C < smallC && x.E[r, c].I != MultiFD.fZero)
							smallC = x.E[r, c].I.C;
					}
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						if (x.E[r, c].C == smallC)
							MultiFD.LopF(x.E[r, c], z.E[r, c]);
						else
							MultiFD.SetToF(x.E[r, c], z.E[r, c]);
						if (x.E[r, c].I.C == smallC)
							MultiFD.LopF(x.E[r, c].I, z.E[r, c].I);
						else
							MultiFD.SetToF(x.E[r, c].I, z.E[r, c].I);
					}
			}
		} // LopZ

		public static void RLopZ(MultiZ z)	// z = z with one less least significant "digit"
		{
			LopZ(z, z);
		}

		public static bool ZTestZ(MultiZ z) // true if z is Zero
		{
			if (z.Rs == 0)
				return (MultiCD.ZTestC(z));
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)				
				{
					if (!MultiCD.ZTestC(z.E[r, c]))
						return false;
				}
			return true;
		}

		public static void MulQ(MultiZ x, MultiZ y, MultiZ z) // z = x * y, all quaternions
		{
			if (NotQ(x) || NotQ(y))
				ErrNotQuat();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ zM = new MultiZ(Calc.fMC, 4, 1);
				zM.E[0, 0] &= x.E[0, 0] * y.E[3, 0] +
					x.E[1, 0] * y.E[2, 0] -
					x.E[2, 0] * y.E[1, 0] +
					x.E[3, 0] * y.E[0, 0];
				zM.E[1, 0] &= -x.E[0, 0] * y.E[2, 0] +
					x.E[1, 0] * y.E[3, 0] +
					x.E[2, 0] * y.E[0, 0] +
					x.E[3, 0] * y.E[1, 0];
				zM.E[2, 0] &= x.E[0, 0] * y.E[1, 0] -
					x.E[1, 0] * y.E[0, 0] +
					x.E[2, 0] * y.E[3, 0] +
					x.E[3, 0] * y.E[2, 0];
				zM.E[3, 0] &= -x.E[0, 0] * y.E[0, 0] -
					x.E[1, 0] * y.E[1, 0] -
					x.E[2, 0] * y.E[2, 0] +
					x.E[3, 0] * y.E[3, 0];
				SetToZ(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MulQ
	
		public static void AugQ(MultiZ x, MultiZ z) // z = quaternion augment of vector x
		{
			if (x.Rs != 3 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				z.E[0, 0] &= x.E[0, 0];
				z.E[1, 0] &= x.E[1, 0];
				z.E[2, 0] &= x.E[2, 0];
				z.E[3, 0] &= MultiFD.fZero;
			}
			if (NotQ(z))
				ErrNotQuat();
		} // AugQ
	
		public static void VPartQ(MultiZ x, MultiZ z) // z = vector part of quaternion x
		{
			if (NotQ(x))
				ErrNotQuat();
			else if (z.Rs != 3 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				z.E[0, 0] &= x.E[0, 0];
				z.E[1, 0] &= x.E[1, 0];
				z.E[2, 0] &= x.E[2, 0];
			}
		} // VPartQ
	
		public static void PrinQ(MultiZ x, MultiZ z)	// z = principal equivalent quaternion of x
		{
			if (NotQ(x) || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				int nZ = 3;
				while (nZ > 0 && x.E[nZ, 0] == MultiFD.fZero)
					nZ--;
				if (x.E[nZ, 0].S)
					EqQ(x, z);
				else
					z &= x;
			}
		} // PrinQ
	
		public static void MToQ(MultiZ x, MultiZ z)	// z = quaternion equivalent of matrix x
		{
			if (NotRZ(x))
				ErrNotRotationZ();
			else if (z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF t0 = new MultiF(Calc.fMC);
				MultiF t1 = new MultiF(Calc.fMC);
				MultiF t2 = new MultiF(Calc.fMC);
				MultiF t3 = new MultiF(Calc.fMC);
				t0 &= MultiFD.fOne + x.E[0, 0] - x.E[1, 1] - x.E[2, 2];
				t1 &= MultiFD.fOne - x.E[0, 0] + x.E[1, 1] - x.E[2, 2];
				t2 &= MultiFD.fOne - x.E[0, 0] - x.E[1, 1] + x.E[2, 2];
				t3 &= MultiFD.fOne + x.E[0, 0] + x.E[1, 1] + x.E[2, 2];
				if (t0.S || t1.S || t2.S || t3.S)
					ErrNotRotationZ();
				else
				{
					int max = 0;
					if (t0 < t1) { max = 1; t0 &= t1; }
					if (t0 < t2) { max = 2; t0 &= t2; }
					if (t0 < t3) { max = 3; t0 &= t3; }
					MultiFD.SqRootF(t0, t0);
					t1 &= t0 / MultiFD.fTwo;
					t2 &= t0 * MultiFD.fTwo;
					if (max == 0)
					{
						z.E[0, 0] &= t1;
						z.E[1, 0] &= (x.E[0, 1] + x.E[1, 0]) / t2;
						z.E[2, 0] &= (x.E[2, 0] + x.E[0, 2]) / t2;
						z.E[3, 0] &= (x.E[1, 2] - x.E[2, 1]) / t2;
					}
					if (max == 1)
					{
						z.E[0, 0] &= (x.E[0, 1] + x.E[1, 0]) / t2;
						z.E[1, 0] &= t1;
						z.E[2, 0] &= (x.E[1, 2] + x.E[2, 1]) / t2;
						z.E[3, 0] &= (x.E[2, 0] - x.E[0, 2]) / t2;
					}
					if (max == 2)
					{
						z.E[0, 0] &= (x.E[2, 0] + x.E[0, 2]) / t2;
						z.E[1, 0] &= (x.E[1, 2] + x.E[2, 1]) / t2;
						z.E[2, 0] &= t1;
						z.E[3, 0] &= (x.E[0, 1] - x.E[1, 0]) / t2;
					}
					if (max == 3)
					{
						z.E[0, 0] &= (x.E[1, 2] - x.E[2, 1]) / t2;
						z.E[1, 0] &= (x.E[2, 0] - x.E[0, 2]) / t2;
						z.E[2, 0] &= (x.E[0, 1] - x.E[1, 0]) / t2;
						z.E[3, 0] &= t1;
					}
					PrinQ(z, z);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MToQ
	
		public static void QToZ(MultiZ x, MultiZ z)	// z = matrix equivalent of quaternion x
		{
			if (NotQ(x))
				ErrNotQuat();
			else if (z.Rs != 3 || z.Cs != 3)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				z.E[0, 0] &= x.E[0, 0] * x.E[0, 0] - x.E[1, 0] * x.E[1, 0] -
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[1, 1] &= -x.E[0, 0] * x.E[0, 0] + x.E[1, 0] * x.E[1, 0] -
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[2, 2] &= -x.E[0, 0] * x.E[0, 0] - x.E[1, 0] * x.E[1, 0] +
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[0, 1] &= (x.E[0, 0] * x.E[1, 0] + x.E[2, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[0, 2] &= (x.E[0, 0] * x.E[2, 0] - x.E[1, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[1, 0] &= (x.E[0, 0] * x.E[1, 0] - x.E[2, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[1, 2] &= (x.E[1, 0] * x.E[2, 0] + x.E[0, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[2, 0] &= (x.E[0, 0] * x.E[2, 0] + x.E[1, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[2, 1] &= (x.E[1, 0] * x.E[2, 0] - x.E[0, 0] * x.E[3, 0]) * MultiFD.fTwo;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // QToM
	
		public static void QToR(MultiZ x, MultiZ z)	// // z = rotation vector from rotation quaternion
		{
			if (NotQ(x))
				ErrNotQuat();
			else if (z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF n3 = new MultiF(Calc.fMC);
				MultiZ a = new MultiZ(Calc.fMC);
				n3 &= x.E[0, 0] * x.E[0, 0] + x.E[1, 0] * x.E[1, 0] + x.E[2, 0] * x.E[2, 0];
				MultiFD.RSqRtF(n3);
				if (n3 > MultiFD.fOne)
					n3 &= MultiFD.fOne;
				int tI;
				Common.InitNAdd(1, "QToR-tI", out tI);
				XPCalLst.ASinN(n3, tI);
				if (x.E[3, 0].S)
				{
					if (Calc.degMode)
					{
						XPCalLst.AddN(180 * MultiFD.fOne, -Common.nList[tI], tI);
					}
					else
					{
						int pI;
						pI = XPCalLst.GetPiPtr();
						XPCalLst.AddN(Common.nList[pI], -Common.nList[tI], tI);
					}
				}
				if (n3 == MultiFD.fZero)
				{
					z.E[0, 0] &= MultiFD.fZero;
					z.E[1, 0] &= MultiFD.fZero;
					z.E[2, 0] &= MultiFD.fZero;
				}
				else
				{
					z.E[0, 0] &= x.E[0, 0] / n3;
					z.E[1, 0] &= x.E[1, 0] / n3;
					z.E[2, 0] &= x.E[2, 0] / n3;
				}
				z.E[3, 0] &= Common.nList[tI] * MultiFD.fTwo;
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // QToR
	
		public static void RToQ(MultiZ x, MultiZ z)	// z = rotation quaternion from rotation vector
		{
			if (NotQ(x))
				ErrNotRotationZ();
			else if (z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF t = new MultiF(Calc.fMC);
				int tI;
				Common.InitNAdd(1, "RToQ-tI", out tI);
				t &= x.E[3, 0] / MultiFD.fTwo;
				XPCalLst.SinN(t, tI);
				z.E[0, 0] &= Common.nList[tI] * x.E[0, 0];
				z.E[1, 0] &= Common.nList[tI] * x.E[1, 0];
				z.E[2, 0] &= Common.nList[tI] * x.E[2, 0];
				XPCalLst.CosN(t, tI);
				z.E[3, 0] &= Common.nList[tI];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // RToQ
	
		public static void MToR(MultiZ x, MultiZ z)	// // z = rotation vector from rotation matrix
		{
			if (NotRZ(x))
				ErrNotRotationZ();
			else if (z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				MToQ(x, z);
				QToR(z, z);
			}
		} // MToR
	
		public static void RToZ(MultiZ x, MultiZ z)	// z = rotation matrix from rotation vector
		{
			if (NotQ(x))
				ErrNotRotationZ();
			if (z.Rs != 3 || z.Cs != 3)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ q = new MultiZ(Calc.fMC, 4, 1);
				RToQ(x, q);
				QToZ(q, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // RToM
	
		public static void RotateQ(MultiZ x, MultiZ y, MultiZ z)
		// z = 3-D vector y rotated by quaternion x
		{
			if (NotQ(x))
				ErrNotQuat();
			else if (y.Rs != 3 || y.Cs != 1 || z.Rs != 3 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				MultiZ yA = new MultiZ(Calc.fMC, 4, 1);
				MultiZ xC = new MultiZ(Calc.fMC, 4, 1);
				AugQ(y, yA);
				InvQ(x, xC);
				MulQ(xC, yA, xC);
				MulQ(xC, x, xC);
				VPartQ(xC, z);
			}
		} // RotateQ
	
		public static void RotateZ(MultiZ x, MultiZ y, MultiZ z)
		// z = 3-D vector y rotated by matrix x
		{
			if (NotRZ(x))
				ErrNotRotationZ();
			if (y.Rs != 3 || y.Cs != 1 || z.Rs != 3 || z.Cs != 1)
				ErrIncompatible();
			else
				MulZ(x, y, z); 
		} // RotateM
	
		public static bool NotQ(MultiZ x) // is x not a quaternion
		{
			if (x.Rs != 4 || x.Cs != 1)
				return true;
			for (int r = 0; r < 4; r++)
				if (x.E[r, 0].I != MultiFD.fZero)
					return true;
			return false;
		} // NotQ
	
		public static bool NotRZ(MultiZ x) // is x not a rotation matrix
		{
			if (x.Rs != 3 || x.Cs != 3)
				return true;
			for (int r = 0; r < 3; r++)
				for (int c = 0; c < 3; c++)
					if (x.E[r, c].I != MultiFD.fZero)
						return true;
			return false;
		} // NotRM
	
		public static void EqQ(MultiZ x, MultiZ z) // z = quaternion equivalent to x
		{
			if (NotQ(x) || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
				z &= -x;
		} // EqQ
	
		public static void InvQ(MultiZ x, MultiZ z) // z = inverse of quaternion x
		{
			EqQ(x, z);
			MultiCD.ChSC(z.E[3, 0]);
		} // InvQ
	
		public static void NormQ(MultiZ x, MultiZ z) // z = quaternion x normalized
		{
			if (NotQ(x) || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ zM = new MultiZ(Calc.fMC, 4, 1);
				MultiZ nX = new MultiZ(Calc.fMC);
				NormV(x, nX);
				if (nX != mZero)
					zM &= x / nX;
				SetToZ(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormQ
	
		public static void NormR(MultiZ x, MultiZ z) // z = rotation vector x normalized
		{
			if (NotQ(x) || z.Rs != 4 || z.Cs != 1)
				ErrNotRotationZ();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ zM = new MultiZ(Calc.fMC, 4, 1);
				MultiZ nX = new MultiZ(Calc.fMC);
				zM &= x;
				zM.E[3, 0] &= mZero;
				NormV(zM, nX);
				if (nX != mZero)
				{
					zM &= zM / nX;
					if (Calc.degMode)
						XPCalLst.Mod(x.E[3, 0], 180 * MultiFD.fTwo, zM.E[3, 0]);
					else
					{
						int pI;
						pI = XPCalLst.GetPiPtr();
						XPCalLst.Mod(x.E[3, 0], Common.nList[pI] * MultiFD.fTwo, zM.E[3, 0]);
					}
				}
				SetToZ(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormR
	
		public static void NormV(MultiZ x, MultiZ z) // z = norm of vector x
		// z = length of vector x = sqrt(x11.r^2 + x11.i^2 + ...
		{
			if (x.Rs != 1 && x.Cs != 1)
				ErrNotVector();
			else if (z.Rs != 0)
				ErrNotScalar();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF zF = new MultiF(Calc.fMC);
				if (x.Rs == 1)
					for (int i = 0; i < x.Cs; i++)
					{
						zF+= (MultiF)x.E[0, i] * (MultiF)x.E[0, i];
						zF+= x.E[0, i].I * x.E[0, i].I;
					}
				else
					for (int i = 0; i < x.Rs; i++)
					{
						zF += (MultiF)x.E[i, 0] * (MultiF)x.E[i, 0];
						zF += x.E[i, 0].I * x.E[i, 0].I;
					}
				MultiFD.RSqRtF(zF);
				SetToFZ(zF, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormV
	
		public static void ErrElementOfScalar()
		{
			MultiID.MuWriteErr("Cannot set an element of a scalar");
		}
	
		public static void ErrNotScalar()
		{
			MultiID.MuWriteErr("Cannot use a matrix as a scalar");
		}
	
		public static void ErrNotMatrix()
		{
			MultiID.MuWriteErr("Cannot use a scalar as a matrix");
		}
	
		public static void ErrOutside()
		{
			MultiID.MuWriteErr("Element outside of matrix");
		}
	
		public static void ErrIncompatible()
		{
			MultiID.MuWriteErr("Matrices are incompatible");
		}
	
		public static void ErrSingular()
		{
			MultiID.MuWriteErr("Matrix is singular");
		}
	
		public static void ErrNotSquare()
		{
			MultiID.MuWriteErr("This is not a square matrix");
		}
	
		public static void ErrNotVector()
		{
			MultiID.MuWriteErr("This is not a vector");
		}
	
		public static void ErrNotColumn()
		{
			MultiID.MuWriteErr("This is not a column vector");
		}
	
		public static void ErrNotQuat()
		{
			MultiID.MuWriteErr("This is not a quaternion");
		}
	
		public static void ErrNotRotationZ()
		{
			MultiID.MuWriteErr("This is not a rotation matrix");
		}
	
		public static MultiZ mZero;		// mZero = 0.0 zero scalar
		public static MultiZ mOne;		// mOne  = 1.0
		public static MultiZ mTwo;		// mTwo  = 2.0
		public static MultiZ vZero;		// vZero = {0.0} zero vector
	}
}
